from ssl import SSLContext

import aiohttp
from ruamel.yaml import YAML
from zabbix_utils import AsyncZabbixAPI  # type: ignore

from zabbixci.assets import Template

yaml = YAML()


class Zabbix:
    zapi: AsyncZabbixAPI
    _client_session = None

    @property
    def api_version(self):
        return self.zapi.version

    def __init__(self, *args, **kwargs):
        if "ssl_context" in kwargs:
            if kwargs["ssl_context"]:
                if not isinstance(kwargs["ssl_context"], SSLContext):
                    raise ValueError("ssl_context must be an instance of SSLContext")

                self._client_session = aiohttp.ClientSession(
                    connector=aiohttp.TCPConnector(ssl=kwargs["ssl_context"])
                )

            del kwargs["ssl_context"]

        self.zapi = AsyncZabbixAPI(*args, **kwargs, client_session=self._client_session)

    def _get_template_group(self, template_group_names: list[str]):
        if self.api_version < 7.0:
            return self.zapi.send_sync_request(
                "hostgroup.get", {"search": {"name": template_group_names}}
            )["result"]
        else:
            return self.zapi.send_sync_request(
                "templategroup.get", {"search": {"name": template_group_names}}
            )["result"]

    def get_templates(
        self, template_group_names: list[str], filter_list: list[str] | None = None
    ):
        ids = self._get_template_group(template_group_names)

        template_group_ids = [group["groupid"] for group in ids]

        if filter_list:
            return self.zapi.send_sync_request(
                "template.get",
                {"groupids": template_group_ids, "filter": {"host": filter_list}},
            )["result"]
        else:
            return self.zapi.send_sync_request(
                "template.get", {"groupids": template_group_ids}
            )["result"]

    def set_template(self, template_id: int, changes: dict):
        return self.zapi.send_sync_request(
            "template.update", {"templateid": template_id, **changes}
        )["result"]

    def export_template_async(self, template_ids: list[int]):
        return self.zapi.send_async_request(
            "configuration.export",
            {"options": {"templates": template_ids}, "format": "yaml"},
        )

    def import_template(self, template: Template):
        export = template.export()

        return self.zapi.send_sync_request(
            "configuration.import",
            {
                "format": "yaml",
                "rules": {
                    **(
                        {
                            "template_groups": {
                                "createMissing": True,
                                "updateExisting": True,
                            }
                        }
                        if self.api_version >= 7.0
                        else {
                            "groups": {
                                "createMissing": True,
                                "updateExisting": True,
                            }
                        }
                    ),
                    "templateLinkage": {"createMissing": True, "deleteMissing": True},
                    "templates": {
                        "createMissing": True,
                        "updateExisting": True,
                    },
                    "discoveryRules": {
                        "createMissing": True,
                        "updateExisting": True,
                        "deleteMissing": True,
                    },
                    "graphs": {
                        "createMissing": True,
                        "updateExisting": True,
                        "deleteMissing": True,
                    },
                    "httptests": {
                        "createMissing": True,
                        "updateExisting": True,
                        "deleteMissing": True,
                    },
                    "items": {
                        "createMissing": True,
                        "updateExisting": True,
                        "deleteMissing": True,
                    },
                    "templateDashboards": {
                        "createMissing": True,
                        "updateExisting": True,
                        "deleteMissing": True,
                    },
                    "triggers": {
                        "createMissing": True,
                        "updateExisting": True,
                        "deleteMissing": True,
                    },
                    "valueMaps": {
                        "createMissing": True,
                        "updateExisting": True,
                        "deleteMissing": True,
                    },
                },
                "source": export,
            },
        )["result"]

    def get_templates_name(self, name: list[str]):
        return self.zapi.send_sync_request("template.get", {"filter": {"host": name}})[
            "result"
        ]

    def delete_templates(self, template_ids: list[int]):
        return self.zapi.send_sync_request("template.delete", template_ids)["result"]

    def get_images(self, search: list[str] | None = None):
        """
        Export images from Zabbix

        TODO: Add batching for large number of images
        """
        if not search:
            return self.zapi.send_sync_request(
                "image.get", {"output": "extend", "select_image": True}
            )["result"]
        else:
            return self.zapi.send_sync_request(
                "image.get",
                {"output": "extend", "select_image": True, "filter": {"name": search}},
            )["result"]

    def create_image(self, image: dict):
        return self.zapi.send_sync_request("image.create", image)["result"]

    def update_image(self, image: dict):
        return self.zapi.send_sync_request("image.update", image)["result"]

    def delete_images(self, image_ids: list[int]):
        return self.zapi.send_sync_request("image.delete", image_ids)["result"]

    def get_icon_maps(self, search: list[str] | None = None):
        """
        Export icon_maps from Zabbix
        """
        if not search:
            return self.zapi.send_sync_request(
                "iconmap.get", {"output": "extend", "selectMappings": "extend"}
            )["result"]
        else:
            return self.zapi.send_sync_request(
                "iconmap.get",
                {
                    "output": "extend",
                    "selectMappings": "extend",
                    "filter": {"name": search},
                },
            )["result"]

    def update_icon_map(self, icon_map: dict):
        return self.zapi.send_sync_request("iconmap.update", icon_map)["result"]

    def create_icon_map(self, icon_map: dict):
        return self.zapi.send_sync_request("iconmap.create", icon_map)["result"]

    def delete_icon_maps(self, icon_map_ids: list[int]):
        return self.zapi.send_sync_request("iconmap.delete", icon_map_ids)["result"]

    def get_server_version(self):
        return self.zapi.send_sync_request("apiinfo.version", need_auth=False)["result"]



import os

from ruamel.yaml import YAML

yaml = YAML()


class Settings:
    _DYN_IMG_EXT = ["png", "jpg", "jpeg", "gif", "bmp", "svg"]
    VERBOSE: bool | None = False
    DEBUG: bool | None = False
    DEBUG_ALL: bool | None = False
    ZABBIX_URL: str | None = "http://localhost:8080"
    ZABBIX_USER: str | None = None
    ZABBIX_PASSWORD: str | None = None
    ZABBIX_TOKEN: str | None = None
    REMOTE: str | None = None
    ROOT_TEMPLATE_GROUP: str = "Templates"
    GIT_AUTHOR_NAME: str = "ZabbixCI"
    GIT_AUTHOR_EMAIL: str = "zabbixci@localhost"
    PULL_BRANCH: str = "main"
    PUSH_BRANCH: str = "main"
    TEMPLATE_PREFIX_PATH: str = "templates"
    IMAGE_PREFIX_PATH: str = "images"
    ICON_MAP_PREFIX_PATH: str = "icon-maps"
    TEMPLATE_WHITELIST: str = ""
    TEMPLATE_BLACKLIST: str = ""
    CACHE_PATH: str = "./cache"
    BATCH_SIZE: int = 5
    IGNORE_TEMPLATE_VERSION: bool = False
    INSECURE_SSL_VERIFY: bool = False
    GIT_USERNAME: str = "git"
    GIT_PASSWORD: str | None = None
    GIT_PUBKEY: str | None = None
    GIT_PRIVKEY: str | None = None
    GIT_KEYPASSPHRASE: str | None = None
    CA_BUNDLE: str | None = None
    DRY_RUN: bool = False
    VENDOR: str | None = None
    SET_VERSION: bool = False
    SYNC_ICONS: bool = False
    SYNC_BACKGROUNDS: bool = False
    SYNC_TEMPLATES: bool = True
    SYNC_ICON_MAPS: bool = False
    IMAGE_WHITELIST: str = ""
    IMAGE_BLACKLIST: str = ""
    ICON_SIZES: str = "24,48,64,128"
    BACKGROUND_SIZES: str = "480,720,1080"
    REGEX_MATCHING: bool = False
    ICON_MAP_WHITELIST: str = ""
    ICON_MAP_BLACKLIST: str = ""

    @classmethod
    def get_template_whitelist(cls):
        return cls.TEMPLATE_WHITELIST.split(",") if cls.TEMPLATE_WHITELIST else []

    @classmethod
    def get_template_blacklist(cls):
        return cls.TEMPLATE_BLACKLIST.split(",") if cls.TEMPLATE_BLACKLIST else []

    @classmethod
    def get_image_whitelist(cls):
        return cls.IMAGE_WHITELIST.split(",") if cls.IMAGE_WHITELIST else []

    @classmethod
    def get_image_blacklist(cls):
        return cls.IMAGE_BLACKLIST.split(",") if cls.IMAGE_BLACKLIST else []

    @classmethod
    def get_icon_map_whitelist(cls):
        return cls.ICON_MAP_WHITELIST.split(",") if cls.ICON_MAP_WHITELIST else []

    @classmethod
    def get_icon_map_blacklist(cls):
        return cls.ICON_MAP_BLACKLIST.split(",") if cls.ICON_MAP_BLACKLIST else []

    @classmethod
    def get_ICON_SIZES(cls):
        size_strings = cls.ICON_SIZES.split(",") if cls.ICON_SIZES else []
        return [int(size) for size in size_strings]

    @classmethod
    def get_BACKGROUND_SIZES(cls):
        size_strings = cls.BACKGROUND_SIZES.split(",") if cls.BACKGROUND_SIZES else []
        return [int(size) for size in size_strings]

    @classmethod
    def from_env(cls):
        for key, value in cls.__dict__.items():
            if key in os.environ:
                if isinstance(value, bool):
                    setattr(cls, key, os.environ[key].lower() == "true")
                else:
                    setattr(cls, key, os.environ[key])

    @classmethod
    def read_config(cls, path):
        with open(path, "r") as f:
            data = yaml.load(f)
            for key, value in data.items():
                setattr(cls, key.upper(), value)



import unittest
from os import getenv

from base_images import BaseImages

from zabbixci.settings import Settings

DEV_ZABBIX_URL = getenv("ZABBIX_URL")
DEV_ZABBIX_TOKEN = getenv("ZABBIX_TOKEN")
DEV_GIT_REMOTE = getenv("REMOTE")


class TestImageWhitelist(BaseImages, unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        super().setUp()
        Settings.IMAGE_WHITELIST = "Cloud_(24),Cloud_(48),Cloud_(64),Cloud_(96),Cloud_(128),Cloud_(128) (renamed),retigra_(200)"



import logging
import os
import ssl
from datetime import datetime, timezone

from pygit2.enums import FileStatus, ResetMode
from regex import search
from ruamel.yaml import YAML

from zabbixci.assets import Template
from zabbixci.cache.cleanup import Cleanup
from zabbixci.git import Git, GitCredentials
from zabbixci.handlers.synchronization.icon_map_synchronization import IconMapHandler
from zabbixci.handlers.synchronization.image_synchronization import ImageHandler
from zabbixci.handlers.synchronization.template_synchronization import TemplateHandler
from zabbixci.settings import Settings
from zabbixci.zabbix import Zabbix


class ZabbixCI:
    logger = logging.getLogger(__name__)
    yaml = YAML()

    _zabbix: Zabbix | None = None
    _git: Git | None = None

    _ssl_context = None

    def __init__(self, settings=Settings, logger=None):
        self._settings = settings

        if logger:
            self.logger = logger

    async def create_zabbix(self) -> None:
        """
        Create a Zabbix object with the appropriate credentials
        """
        # Construct the SSL context if a CA bundle is provided

        if Settings.CA_BUNDLE:
            self._ssl_context = ssl.create_default_context()
            self._ssl_context.load_verify_locations(Settings.CA_BUNDLE)

        self._zabbix = Zabbix(
            url=self._settings.ZABBIX_URL,
            validate_certs=not self._settings.INSECURE_SSL_VERIFY,
            ssl_context=self._ssl_context,
        )

        if self._settings.ZABBIX_USER and self._settings.ZABBIX_PASSWORD:
            self.logger.debug("Using username and password for Zabbix authentication")
            await self._zabbix.zapi.login(
                user=self._settings.ZABBIX_USER, password=self._settings.ZABBIX_PASSWORD
            )
        elif self._settings.ZABBIX_TOKEN:
            self.logger.debug("Using token for Zabbix authentication")
            await self._zabbix.zapi.login(token=self._settings.ZABBIX_TOKEN)

        if self._zabbix.zapi.version < 6.0:
            self.logger.error(
                f"Zabbix server version {self._zabbix.zapi.version} is not supported (7.0+ required)"
            )
            raise SystemExit(1)

    def create_git(self, git_cb=None) -> None:
        """
        Create a Git object with the appropriate credentials
        """
        if git_cb is None:
            git_cb = GitCredentials().create_git_callback()

        self._git = Git(self._settings.CACHE_PATH, git_cb)

    async def push(self) -> bool:
        """
        Fetch Zabbix state and commit changes to git remote
        """
        if not self._git or not self._zabbix:
            raise ValueError(
                "ZabbixCI has not been initialized, call create_zabbix() and create_git() first"
            )

        if not Settings.REMOTE:
            raise ValueError("Remote repository not set")

        self._git.fetch(Settings.REMOTE)

        if not self._git.is_empty:
            # If the repository is empty, new branches can't be created. But it is
            # safe to push to the default branch
            self._git.switch_branch(Settings.PUSH_BRANCH)

            # Pull the latest remote state
            try:
                self._git.pull(Settings.REMOTE)
            except KeyError:
                self.logger.info(
                    f"Remote branch does not exist, using state from branch: {Settings.PULL_BRANCH}"
                )
                # Remote branch does not exist, we pull the default branch and create a new branch
                self._git.switch_branch(Settings.PULL_BRANCH)
                self._git.pull(Settings.REMOTE)

                # Create a new branch
                self._git.switch_branch(Settings.PUSH_BRANCH)

        # Reflect current Zabbix state in the cache
        Cleanup.cleanup_cache()

        template_handler = TemplateHandler(self._zabbix)
        image_handler = ImageHandler(self._zabbix)
        icon_map_handler = IconMapHandler(self._zabbix)

        template_objects = await template_handler.templates_to_cache()
        image_objects = image_handler.images_to_cache()
        icon_map_handler.icon_map_to_cache(image_objects)

        # Check if there are any changes to commit
        if not self._git.has_changes and not self._git.ahead_of_remote:
            self.logger.info("No changes detected")
            return False

        self.logger.info("Remote differs from local state, preparing to push")
        change_amount = len(self._git.status())

        diff = self._git.diff()
        Git.print_diff(diff)

        # Check if we have any changes to commit. Otherwise, we just push the current state
        if self._git.has_changes:
            # Create a commit
            changes = self._git.status()

            regex_match = search("https?://([^/]+)", self._zabbix.zapi.url)

            host = os.getenv(
                "ZABBIX_HOST",
                regex_match.group(1) if regex_match else "zabbix_host",
            )

            change_amount = len(changes)

            for relative_path, status in changes.items():
                file = f"{Settings.CACHE_PATH}/{relative_path}"

                if status == FileStatus.WT_DELETED:
                    self.logger.info(f"Detected deletion of: {file}")
                    continue

                self.logger.info(f"Detected change in: {file}")

                if not template_handler.read_validation(file):
                    continue

                template = Template.open(file)

                if (
                    Settings.VENDOR
                    and not template.vendor
                    and self._zabbix.api_version >= 7.0
                ):
                    set_vendor = Settings.VENDOR
                    template.set_vendor(set_vendor)
                    self.logger.debug(f"Setting vendor to: {set_vendor}")

                if Settings.SET_VERSION and self._zabbix.api_version >= 7.0:
                    new_version = datetime.now(timezone.utc).strftime("%Y.%m.%d %H:%M")
                    template.set_version(new_version)
                    self.logger.debug(f"Setting version to: {new_version}")

                if (template.new_version or template.new_vendor) and (
                    template.vendor and template.version
                ):
                    template.save()

                    if not Settings.DRY_RUN:
                        self.logger.debug(
                            f"Updating template metadata for: {template.name}"
                        )
                        self._zabbix.set_template(
                            next(
                                filter(
                                    lambda t: t["host"] == template.name,
                                    template_objects,
                                )
                            )["templateid"],
                            template.updated_items,
                        )

            # Commit and push the changes
            self._git.add_all()

            if not Settings.DRY_RUN:
                # Generate commit message
                self._git.commit(f"Committed Zabbix state from {host}")
                self.logger.info(
                    f"Staged changes from {host} committed to {self._git.current_branch}"
                )
        else:
            self.logger.info("No staged changes, updating remote with current state")

        if not self._settings.DRY_RUN:
            self._git.push(Settings.REMOTE)
            self.logger.info(
                f"Committed {change_amount} new changes to {Settings.REMOTE}:{Settings.PUSH_BRANCH}"
            )
        else:
            self.logger.info(
                f"Dry run enabled, would have committed {change_amount} new changes to {Settings.REMOTE}:{Settings.PUSH_BRANCH}"
            )

        return change_amount > 0

    async def pull(self) -> bool:
        """
        Pull current state from git remote and update Zabbix
        """
        if not self._git or not self._zabbix:
            raise ValueError(
                "ZabbixCI has not been initialized, call create_zabbix() and create_git() first"
            )

        if not Settings.REMOTE:
            raise ValueError("Remote repository not set")

        self._git.switch_branch(Settings.PULL_BRANCH)

        # Pull the latest remote state, untracked changes are preserved
        self._git.pull(Settings.REMOTE)

        self._git.reset(
            self._git.lookup_reference(
                f"refs/remotes/origin/{Settings.PULL_BRANCH}"
            ).target,
            ResetMode.HARD,
        )

        current_revision = self._git.get_current_revision()

        # Reflect current Zabbix state in the cache
        Cleanup.cleanup_cache()

        template_handler = TemplateHandler(self._zabbix)
        image_handler = ImageHandler(self._zabbix)
        icon_map_handler = IconMapHandler(self._zabbix)

        template_objects = await template_handler.templates_to_cache()
        image_objects = image_handler.images_to_cache()
        icon_map_objects = icon_map_handler.icon_map_to_cache(image_objects)

        # Check if there are any changes to commit
        if self._git.has_changes:
            self.logger.info("Zabbix state is out of sync, syncing")

        status = self._git.status()

        # Get the changed files, we compare the untracked changes with the desired.
        # When we have a new untracked file, that means it was deleted in the desired state.
        changed_files: list[str] = [
            f"{Settings.CACHE_PATH}/{path}"
            for path, flags in status.items()
            if flags in [FileStatus.WT_DELETED, FileStatus.WT_MODIFIED]
        ]
        deleted_files: list[str] = [
            f"{Settings.CACHE_PATH}/{path}"
            for path, flags in status.items()
            if flags == FileStatus.WT_NEW
        ]

        self.logger.debug(f"Following files have changed on Git: {changed_files}")
        self.logger.debug(f"Following files are deleted from Git: {deleted_files}")

        diff = self._git.diff()
        Git.print_diff(diff, invert=True)

        # Sync the file cache with the desired git state
        self._git.reset(current_revision, ResetMode.HARD)

        imported_template_ids = template_handler.import_file_changes(changed_files)
        deleted_template_names = template_handler.delete_file_changes(
            deleted_files, imported_template_ids, template_objects
        )

        imported_images = image_handler.import_file_changes(
            changed_files, image_objects
        )

        if len(imported_images) > 0:
            # Update available image objects (only needed for Zabbix image id's)
            image_objects = image_handler.images_to_cache()

        imported_icon_maps = icon_map_handler.import_file_changes(
            changed_files, icon_map_objects, image_objects
        )
        deleted_icon_map_names = icon_map_handler.delete_file_changes(
            deleted_files, imported_icon_maps, icon_map_objects
        )

        deleted_image_names = image_handler.delete_file_changes(
            deleted_files, imported_images, image_objects
        )

        # Inform user about the changes
        if Settings.DRY_RUN:
            self.logger.info(
                "Dry run enabled, no changes will be made to Zabbix. "
                f"Would have imported {len(imported_template_ids)} templates, deleted {len(deleted_template_names)} templates. "
                f"Would have imported {len(imported_images)} images, deleted {len(deleted_image_names)} images. "
                f"Would have imported {len(imported_icon_maps)} icon maps, deleted {len(deleted_icon_map_names)} icon maps."
            )
        else:
            if (
                len(deleted_template_names) == 0
                and len(imported_template_ids) == 0
                and len(imported_images) == 0
                and len(deleted_image_names) == 0
                and len(imported_icon_maps) == 0
            ):
                self.logger.info("No changes detected, Zabbix is up to date")
            else:
                self.logger.info(
                    "Zabbix state has been synchronized. "
                    f"Imported {len(imported_template_ids)} templates, deleted {len(deleted_template_names)} templates. "
                    f"Imported {len(imported_images)} images, deleted {len(deleted_image_names)} images. "
                    f"Imported {len(imported_icon_maps)} icon maps, deleted {len(deleted_icon_map_names)} icon maps."
                )

        # clean local changes
        self._git.clean()
        return (
            len(imported_template_ids) > 0
            or len(deleted_template_names) > 0
            or len(imported_images) > 0
            or len(deleted_image_names) > 0
            or len(imported_icon_maps) > 0
            or len(deleted_icon_map_names) > 0
        )

    def generate_images(self, type: str) -> bool:
        """
        Generate icons/backgrounds from Zabbix and save them to the cache
        """
        if not self._git or not self._zabbix:
            raise ValueError(
                "ZabbixCI has not been initialized, call create_zabbix() and create_git() first"
            )

        if not Settings.REMOTE:
            raise ValueError("Remote repository not set")

        self._git.fetch(Settings.REMOTE)

        self.logger.info("Generating icons from Zabbix")

        image_handler = ImageHandler(self._zabbix)

        if type == "background":
            image_handler.generate_backgrounds()
        elif type == "icon":
            image_handler.generate_icons()

        if not self._git.is_empty:
            # If the repository is empty, new branches can't be created. But it is
            # safe to push to the default branch
            self._git.switch_branch(Settings.PUSH_BRANCH)

            # Pull the latest remote state
            try:
                self._git.pull(Settings.REMOTE)
            except KeyError:
                self.logger.info(
                    f"Remote branch does not exist, using state from branch: {Settings.PULL_BRANCH}"
                )
                # Remote branch does not exist, we pull the default branch and create a new branch
                self._git.switch_branch(Settings.PULL_BRANCH)
                self._git.pull(Settings.REMOTE)

                # Create a new branch
                self._git.switch_branch(Settings.PUSH_BRANCH)

        # Check if there are any changes to commit
        if not self._git.has_changes and not self._git.ahead_of_remote:
            self.logger.info("No changes detected")
            return False

        self.logger.info("Remote differs from local state, preparing to push")
        change_amount = len(self._git.status())

        # Check if we have any changes to commit. Otherwise, we just push the current state
        if self._git.has_changes:
            # Commit and push the changes
            self._git.add_all()

            if not Settings.DRY_RUN:
                # Generate commit message
                self._git.commit(f"Generated {type}(s) from source files")
                self.logger.info(
                    f"Staged changes from generated images committed to {self._git.current_branch}"
                )
        else:
            self.logger.info("No staged changes, updating remote with current state")

        if not self._settings.DRY_RUN:
            self._git.push(Settings.REMOTE)
            self.logger.info(
                f"Committed {change_amount} new changes to {Settings.REMOTE}:{Settings.PUSH_BRANCH}"
            )
        else:
            self.logger.info(
                f"Dry run enabled, would have committed {change_amount} new changes to {Settings.REMOTE}:{Settings.PUSH_BRANCH}"
            )

        return change_amount > 0



import logging

from zabbixci.assets.icon_map import IconMap
from zabbixci.assets.image import Image
from zabbixci.handlers.validation.icon_map_validation import IconMapValidationHandler
from zabbixci.settings import Settings
from zabbixci.zabbix.zabbix import Zabbix

logger = logging.getLogger(__name__)


class IconMapHandler(IconMapValidationHandler):
    """
    Handler for importing icon maps into Zabbix based on changed files. Includes validation steps based on settings.

    :param zabbix: Zabbix instance
    """

    _zabbix: Zabbix

    def __init__(self, zabbix: Zabbix):
        self._zabbix = zabbix

    def icon_map_to_cache(self, images: list[Image]) -> list[IconMap]:
        """
        Export Zabbix icon maps to cache.
        """
        if not Settings.SYNC_ICON_MAPS:
            return []

        if not Settings.SYNC_ICONS:
            logger.warning(
                "SYNC_ICONS is disabled, unable to export icon maps without icons"
            )
            return []

        search = (
            self.get_whitelist()
            if not self._use_regex() and self.get_whitelist()
            else None
        )

        icon_maps = self._zabbix.get_icon_maps(search)

        logger.info(f"Found {len(icon_maps)} icon map(s) in Zabbix")

        icon_map_objects = []

        for icon_map in icon_maps:
            icon_map_object = IconMap.from_zabbix(icon_map, images)

            if not self.object_validation(icon_map_object):
                continue

            icon_map_object.save()
            icon_map_objects.append(icon_map_object)

        return icon_map_objects

    def import_file_changes(
        self,
        changed_files: list[str],
        icon_map_objects: list[IconMap],
        image_objects: list[Image],
    ) -> list[str]:
        """
        Import icon maps into Zabbix based on changed files.

        :param changed_files: List of changed files
        :param icon_map_objects: List of icon map objects from Zabbix, needed for choice between creation or update
        :param image_objects: List of image objects from Zabbix, needed to open icon_map exports

        :return: List of imported icon_map names
        """
        icon_maps: list[IconMap] = []

        if not Settings.SYNC_ICON_MAPS:
            return []

        for file in changed_files:
            if not self.read_validation(file):
                continue

            icon_map = IconMap.open(file, image_objects)

            if not self.object_validation(icon_map):
                continue

            icon_maps.append(icon_map)
            logger.info(f"Detected change in image: {icon_map.name}")

        def __import_icon_map(icon_map: IconMap):
            if icon_map.name in [t.name for t in icon_map_objects]:
                logger.info(f"Updating: {icon_map.name}")

                old_icon_map = next(
                    filter(lambda dt: dt.name == icon_map.name, icon_map_objects)
                )

                return self._zabbix.update_icon_map(
                    {
                        "iconmapid": old_icon_map.icon_mapid,
                        **icon_map.zabbix_dict,
                    }
                )
            else:
                logger.info(f"Creating: {icon_map.name}")
                return self._zabbix.create_icon_map(icon_map.zabbix_dict)

        failed_icon_maps: list[IconMap] = []

        # Import the images
        for icon_map in icon_maps:
            if not Settings.DRY_RUN:
                try:
                    __import_icon_map(icon_map)
                except Exception as e:
                    logger.warning(
                        f"Error importing icon mapping {icon_map.name}, will try to import later"
                    )
                    logger.debug(f"Error details: {e}")
                    failed_icon_maps.append(icon_map)

        if len(failed_icon_maps):
            for icon_map in failed_icon_maps:
                try:
                    __import_icon_map(icon_map)
                except Exception as e:
                    logger.error(
                        f"Error importing icon mapping {icon_map.name}", exc_info=e
                    )

        return [t.name for t in icon_maps]

    def delete_file_changes(
        self,
        deleted_files: list[str],
        imported_icon_map_names: list[str],
        icon_map_objects: list[IconMap],
    ):
        """
        Delete icon maps from Zabbix based on deleted files.

        :param deleted_files: List of deleted files
        :param imported_icon_map_names: List of imported icon_map names
        :param icon_map_objects: List of icon_map objects from Zabbix, needed for deletion

        :return: List of deleted image names
        """
        deletion_queue: list[str] = []

        if not Settings.SYNC_ICON_MAPS:
            return []

        for file in deleted_files:
            if not self.read_validation(file):
                continue

            icon_map = IconMap.partial_open(file)

            if not icon_map:
                logger.warning(f"Could not open to be deleted file: {file}")
                continue

            if not self.object_validation(icon_map):
                continue

            if icon_map.name in imported_icon_map_names:
                logger.debug(
                    f"Icon map {icon_map.name} was just imported, skipping deletion"
                )
                continue

            deletion_queue.append(icon_map.name)
            logger.info(f"Added {icon_map.name} to deletion queue")

        if len(deletion_queue):
            icon_map_ids = [
                t.icon_mapid
                for t in list(
                    filter(lambda dt: dt.name in deletion_queue, icon_map_objects)
                )
            ]

            logger.info(f"Deleting {len(icon_map_ids)} icon map(s) from Zabbix")

            if len(icon_map_ids):
                if not Settings.DRY_RUN:
                    self._zabbix.delete_icon_maps(icon_map_ids)

        return deletion_queue



import unittest
from os import getenv

from base_images import BaseImages

from zabbixci.settings import Settings

DEV_ZABBIX_URL = getenv("ZABBIX_URL")
DEV_ZABBIX_TOKEN = getenv("ZABBIX_TOKEN")
DEV_GIT_REMOTE = getenv("REMOTE")


class TestImagesBlacklist(BaseImages, unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        super().setUp()
        Settings.REGEX_MATCHING = True
        Settings.IMAGE_BLACKLIST = "retigra.*"

    async def test_image_delete(self):
        image_id = self.zci._zabbix.get_images(["retigra_(200)"])[0]["imageid"]
        self.zci._zabbix.delete_images([image_id])

        changed = await self.zci.push()
        self.assertFalse(changed, "Blacklisted image was deleted")

        changed = await self.zci.pull()
        self.assertFalse(changed, "Blacklisted image was imported")



import logging
import os

from zabbixci.assets.icon_map import IconMap
from zabbixci.assets.image import Image
from zabbixci.assets.template import Template
from zabbixci.handlers.validation.icon_map_validation import IconMapValidationHandler
from zabbixci.handlers.validation.image_validation import ImageValidationHandler
from zabbixci.handlers.validation.template_validation import TemplateValidationHandler
from zabbixci.settings import Settings

logger = logging.getLogger(__name__)


class Cleanup:
    @classmethod
    def match_template_cleanup(cls, root: str, name: str):
        if not Settings.SYNC_TEMPLATES:
            return False

        template_handler = TemplateValidationHandler()
        file = os.path.join(root, name)

        if not template_handler.read_validation(file):
            return False

        template = Template.open(file)

        if not template or not template.is_template:
            logger.warning(f"Could not open file {file} as a template")
            return False

        if not template_handler.object_validation(template):
            return False

        return True

    @classmethod
    def match_image_cleanup(cls, root: str, name: str):
        """
        Check if a file is an image file that should be cleaned up
        """
        if not Settings.SYNC_ICONS and not Settings.SYNC_BACKGROUNDS:
            return False

        image_handler = ImageValidationHandler()

        file = os.path.join(root, name)

        if not image_handler.read_validation(file):
            return False

        image = Image.open(file)

        if not image:
            return False

        if not image_handler.object_validation(image):
            return False

        return True

    @classmethod
    def match_icon_map_cleanup(cls, root: str, name: str):
        """
        Check if a file is an icon_map file that should be cleaned up
        """
        if not Settings.SYNC_ICON_MAPS:
            return False

        icon_map_handler = IconMapValidationHandler()

        file = os.path.join(root, name)

        if not icon_map_handler.read_validation(file):
            return False

        icon_map = IconMap.partial_open(file)

        if not icon_map_handler:
            return False

        if not icon_map_handler.object_validation(icon_map):
            return False

        return True

    @classmethod
    def cleanup_cache(cls, full: bool = False) -> None:
        """
        Clean all .yaml (template) files from the cache directory

        If full is True, also remove the .git directory and all other files
        """
        for root, dirs, files in os.walk(Settings.CACHE_PATH, topdown=False):
            if f"{Settings.CACHE_PATH}/.git" in root and not full:
                continue

            for name in files:
                if (
                    full
                    or cls.match_template_cleanup(root, name)
                    or cls.match_image_cleanup(root, name)
                    or cls.match_icon_map_cleanup(root, name)
                ):
                    os.remove(os.path.join(root, name))

            for name in dirs:
                if name == ".git" and root == Settings.CACHE_PATH and not full:
                    continue

                # Remove empty directories
                if not os.listdir(os.path.join(root, name)):
                    os.rmdir(os.path.join(root, name))

        if full and os.path.exists(Settings.CACHE_PATH):
            os.rmdir(Settings.CACHE_PATH)
            logger.info("Cache directory cleared")



import logging
from os import environ

from zabbixci.settings import Settings

grey = "\x1b[38;20m"
yellow = "\x1b[33;20m"
red = "\x1b[31;20m"
bold_red = "\x1b[31;1m"
reset = "\x1b[0m"


class CustomFormatter(logging.Formatter):

    log_format = "%(asctime)s  [%(levelname)s]: %(message)s"
    formats: dict[str, str] = {}

    def __init__(self):
        if Settings.DEBUG or Settings.DEBUG_ALL:
            self.log_format = "%(asctime)s [%(name)s]  [%(levelname)s]: %(message)s"

        term = environ.get("TERM", "dumb")
        no_color = environ.get("NO_COLOR", "false")

        if term == "dumb" or no_color.lower() == "true":
            self.formats = {
                logging.DEBUG: self.log_format,
                logging.INFO: self.log_format,
                logging.WARNING: self.log_format,
                logging.ERROR: self.log_format,
                logging.CRITICAL: self.log_format,
            }
        else:
            self.formats = {
                logging.DEBUG: grey + self.log_format + reset,
                logging.INFO: grey + self.log_format + reset,
                logging.WARNING: yellow + self.log_format + reset,
                logging.ERROR: red + self.log_format + reset,
                logging.CRITICAL: bold_red + self.log_format + reset,
            }

    def format(self, record):
        log_fmt = self.formats.get(record.levelno)
        formatter = logging.Formatter(log_fmt, "%Y-%m-%d %H:%M:%S")
        return formatter.format(record)



import logging
from base64 import b64decode, b64encode

import regex

from zabbixci.assets.asset import Asset
from zabbixci.cache.cache import Cache
from zabbixci.settings import Settings

logger = logging.getLogger(__name__)


class Image(Asset):
    image_id: int | None = None
    image: bytes
    name: str
    type: str

    def __init__(
        self, base64: str, name: str, type: str = "icon", image_id: int | None = None
    ):
        self.image = b64decode(base64)
        self.name = name
        self.type = type
        self.image_id = image_id

    def __str__(self):
        return f"{self.name} ({self.type})"

    @property
    def _type_folder(self):
        return "icons" if self.type == "icon" else "backgrounds"

    def save(self):
        name_folders = self.name.split("/")[0:-1]

        Cache.makedirs(
            f"{Settings.CACHE_PATH}/{Settings.IMAGE_PREFIX_PATH}/{self._type_folder}/{'/'.join(name_folders)}",
        )

        with Cache.open(
            f"{Settings.CACHE_PATH}/{Settings.IMAGE_PREFIX_PATH}/{self._type_folder}/{self.name}.png",
            "wb",
        ) as file:
            file.write(self.image)

    def load(self, path: str):
        with Cache.open(f"{Settings.CACHE_PATH}/{path}/{self.name}.png", "rb") as file:
            self.image = file.read()

    def as_zabbix_dict(self):
        return {
            "image": b64encode(self.image).decode(),
            "name": self.name,
            "imagetype": "1" if self.type == "icon" else "2",
        }

    def minify(self):
        return Image("", self.name, self.type, self.image_id)

    @classmethod
    def from_zabbix(cls, image: dict):
        """
        Create an Image object from a Zabbix API image object

        See: https://www.zabbix.com/documentation/7.0/en/manual/api/reference/image/object
        """
        return cls(
            image["image"],
            image["name"],
            "icon" if image["imagetype"] == "1" else "background",
            image["imageid"],
        )

    @classmethod
    def open(cls, path: str):
        with Cache.open(path, "rb") as file:
            # TODO: Validation of path ending with /
            matches = regex.match(
                f".*{Settings.IMAGE_PREFIX_PATH}/(icons|backgrounds)/(.*).png", path
            )

            if not matches:
                logger.debug(f"Skipping invalid image path: {path}")
                return None

            type = "icon" if matches[1] == "icons" else "background"
            return cls(b64encode(file.read()).decode(), matches[2], type)



import logging
import os
import os.path
from typing import Self

from zabbixci.cache.filesystem import Filesystem


class Cache(Filesystem):
    _logger = None
    _instance: Self = None
    _cache_dir: str = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(Cache, cls).__new__(cls)
            cls._instance.cache = {}
        return cls._instance

    def __init__(self, cache_dir: str):
        self._cache_dir = os.path.realpath(cache_dir)
        self._logger = logging.getLogger(__name__)

    @classmethod
    def _ensure_safe_path(cls, path):
        if not cls.is_within_cache(path):
            raise ValueError(f"Path {path} does not reside within the cache directory")

    @classmethod
    def get_files(cls, path: str) -> list[str]:
        """
        Wrapped get_files function that ensures that the path is within the cache directory
        """

        cls._ensure_safe_path(path)
        real_path = os.path.realpath(path)

        found_files = []

        for root, _dirs, files in os.walk(real_path):
            for name in files:
                found_files.append(os.path.join(root, name))

        return found_files

    @classmethod
    def open(cls, path, mode):
        """
        Wrapped open function that ensures that the path is within the cache directory
        """
        # Binary operations do not take an encoding argument
        if "rb" in mode or "wb" in mode:
            cls._ensure_safe_path(os.path.dirname(path))
            return open(path, mode)

        cls._ensure_safe_path(path)
        return open(path, mode, encoding="utf-8")

    @classmethod
    def makedirs(cls, path):
        """
        Wrapped makedirs function that ensures that the path is within the cache directory
        """

        cls._ensure_safe_path(path)
        os.makedirs(path, exist_ok=True)

    @classmethod
    def exists(cls, path):
        """
        Wrapped exists function that ensures that the path is within the cache directory
        """

        cls._ensure_safe_path(path)
        return os.path.exists(path)

    @classmethod
    def is_within_cache(cls, path: str):
        """
        Check if a path is in the cache directory
        """
        return cls.is_within(path, cls._instance._cache_dir)



import unittest
from os import getenv

from base_templates import BaseTemplates

from zabbixci.settings import Settings

DEV_ZABBIX_URL = getenv("ZABBIX_URL")
DEV_ZABBIX_TOKEN = getenv("ZABBIX_TOKEN")
DEV_GIT_REMOTE = getenv("REMOTE")


class TestTemplatesBlacklistRegex(BaseTemplates, unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        super().setUp()
        Settings.REGEX_MATCHING = True
        Settings.TEMPLATE_BLACKLIST = "Non .*"

    # Test template deletion with whitelist checks
    async def test_template_delete(self):
        # Delete a template
        template_id = self.zci._zabbix.get_templates_name(
            ["Acronis Cyber Protect Cloud by HTTP"]
        )[0]["templateid"]
        self.zci._zabbix.delete_templates([template_id])

        Settings.TEMPLATE_BLACKLIST = r"(Acronis Cyber Protect \w+ \w+ \w+)"

        # Push changes to git
        changed = await self.zci.push()
        self.assertFalse(changed, "Template deletion detected inside of blacklist")


if __name__ == "__main__":
    unittest.main()



class BaseZabbixCIException(Exception):
    pass


class GitException(BaseZabbixCIException):
    pass


class ZabbixException(BaseZabbixCIException):
    pass


class ZabbixIconMissingException(ZabbixException):
    pass



import logging

from zabbixci.assets.icon_map import IconMap
from zabbixci.cache.filesystem import Filesystem
from zabbixci.handlers.validation.validation_handler import Handler
from zabbixci.settings import Settings

logger = logging.getLogger(__name__)


class IconMapValidationHandler(Handler):
    """
    Handler for importing icon maps into Zabbix based on changed files. Includes validation steps based on settings.
    """

    def get_whitelist(self):
        return Settings.get_icon_map_whitelist()

    def get_blacklist(self):
        return Settings.get_icon_map_blacklist()

    def read_validation(self, changed_file: str) -> bool:
        """
        Validation steps to perform on a changed file before it is processed as a icon_map.
        """
        if not changed_file.endswith(".yaml"):
            return False

        # Check if file is within the desired path
        if not Filesystem.is_within(
            changed_file, f"{Settings.CACHE_PATH}/{Settings.ICON_MAP_PREFIX_PATH}"
        ):
            logger.debug(
                f"Skipping .yaml file {changed_file} outside of icon_map prefix path"
            )
            return False

        return True

    def object_validation(self, icon_map: IconMap | None) -> bool:
        if not icon_map:
            return False

        if self.enforce_whitelist(icon_map.name):
            logger.debug(f"Skipping icon_map {icon_map.name} not in whitelist")
            return False

        if self.enforce_blacklist(icon_map.name):
            logger.debug(f"Skipping icon_map {icon_map.name} in blacklist")
            return False

        return True



import asyncio
import logging
from io import StringIO

from ruamel.yaml import YAML

from zabbixci.assets.template import Template
from zabbixci.handlers.validation.template_validation import TemplateValidationHandler
from zabbixci.settings import Settings
from zabbixci.zabbix.zabbix import Zabbix

logger = logging.getLogger(__name__)
yaml = YAML()


class TemplateHandler(TemplateValidationHandler):
    """
    Handler for importing templates into Zabbix based on changed files. Includes validation steps based on settings.

    :param zabbix: Zabbix instance
    """

    _zabbix: Zabbix

    def __init__(self, zabbix: Zabbix):
        self._zabbix = zabbix

    async def zabbix_export(self, templates: list[dict]):
        batches = [
            templates[i : i + Settings.BATCH_SIZE]
            for i in range(0, len(templates), Settings.BATCH_SIZE)
        ]

        failed_exports = []

        for batchIndex, batch in enumerate(batches):
            logger.info(f"Processing batch {batchIndex + 1}/{len(batches)}")
            coros = []
            for t in batch:
                coros.append(self._zabbix.export_template_async([t["templateid"]]))

            responses = await asyncio.gather(*coros, return_exceptions=True)

            for index, response in enumerate(responses):
                if isinstance(response, BaseException):
                    logger.error(f"Error exporting template: {response}")

                    # Retry the export
                    failed_exports.append(batch[index])
                    continue

                export_yaml = yaml.load(StringIO(response["result"]))

                if "templates" not in export_yaml["zabbix_export"]:
                    logger.info("No templates found in Zabbix")
                    return

                zabbix_template = Template.from_zabbix(export_yaml["zabbix_export"])

                if not self.object_validation(zabbix_template):
                    continue

                zabbix_template.save()
                logger.info(f"Exported Zabbix template: {zabbix_template.name}")

        if len(failed_exports):
            logger.warning(
                f"Failed to export {len(failed_exports)} templates, retrying"
            )
            await self.zabbix_export(failed_exports)

    async def templates_to_cache(self) -> list[dict]:
        """
        Export Zabbix templates to the cache
        """
        if not Settings.SYNC_TEMPLATES:
            return []

        search = (
            self.get_whitelist()
            if not self._use_regex() and self.get_whitelist()
            else None
        )
        templates = self._zabbix.get_templates([Settings.ROOT_TEMPLATE_GROUP], search)

        logger.info(f"Found {len(templates)} templates in Zabbix")
        logger.debug(f"Found Zabbix templates: {[t['host'] for t in templates]}")

        await self.zabbix_export(templates)
        return templates

    def object_validation(self, template) -> bool:
        """
        Validation steps to perform on a template before it is imported into Zabbix
        """
        if not super().object_validation(template):
            return False

        zabbix_version = self._zabbix.get_server_version()

        if (
            not Settings.IGNORE_TEMPLATE_VERSION
            and template.zabbix_version.split(".")[0:2]
            != zabbix_version.split(".")[0:2]
        ):
            logger.warning(
                f"Template {template.name}: {template.zabbix_version} must match Zabbix release {'.'.join(zabbix_version.split('.')[0:2])}"
            )
            return False

        return True

    def import_file_changes(self, changed_files: list[str]) -> list[str]:
        """
        Import templates into Zabbix based on changed files.
        Changes are parsed and validated before importing.

        :param changed_files: List of changed files

        :return: List of changed template UUIDs
        """
        templates: list[Template] = []

        if Settings.SYNC_TEMPLATES is False:
            return []

        for file in changed_files:
            if not self.read_validation(file):
                continue

            template = Template.open(file)

            if not template or not template.is_template:
                logger.warning(f"Could load file {file} as a template")
                continue

            if not self.object_validation(template):
                continue

            templates.append(template)
            logger.info(f"Detected change in template: {template.name}")

        # Group templates by level
        templates = sorted(templates, key=lambda tl: tl.level(templates))

        failed_templates: list[Template] = []

        # Import the templates
        for template in templates:
            if not Settings.DRY_RUN:
                try:
                    logger.info(
                        f"Importing {template.name}, level {template.level(templates)}"
                    )
                    self._zabbix.import_template(template)
                except Exception as e:
                    logger.warning(
                        f"Error importing template {template.name}, will try to import later"
                    )
                    logger.debug(f"Error details: {e}")
                    failed_templates.append(template)

        if len(failed_templates):
            for template in failed_templates:
                try:
                    logger.info(
                        f"Importing {template.name}, level {template.level(templates)} after previous failure"
                    )
                    self._zabbix.import_template(template)
                except Exception as e:
                    logger.error(f"Error importing template {template}: {e}")

        return [
            template_id
            for template in templates
            for template_id in template.template_ids
        ]

    def delete_file_changes(
        self,
        deleted_files: list[str],
        imported_template_ids: list[str],
        template_objects: list[dict],
    ):
        """
        Delete templates from Zabbix based on deleted files.

        :param deleted_files: List of deleted files
        :param imported_template_ids: List of imported template UUIDs
        :param template_objects: List of template objects from Zabbix needed for deletion in current Zabbix instance

        :return: List of deleted template names
        """
        if Settings.SYNC_TEMPLATES is False:
            return []

        deletion_queue: list[str] = []

        # Check if deleted files are templates and if they are imported, if not add to deletion queue
        for file in deleted_files:
            if not self.read_validation(file):
                continue

            template = Template.open(file)

            if not template or not template.is_template:
                logger.warning(f"Could not open to be deleted file: {file}")
                continue

            if not self.object_validation(template):
                continue

            if template.uuid in imported_template_ids:
                logger.debug(
                    f"Template {template.name} is being imported under a different name or path, skipping deletion"
                )
                continue

            deletion_queue.append(template.name)
            logger.info(f"Added {template.name} to deletion queue")

        # Delete templates in deletion queue
        if len(deletion_queue):
            template_ids = [
                # Get template IDs from Zabbix
                t["templateid"]
                for t in list(
                    filter(lambda dt: dt["host"] in deletion_queue, template_objects)
                )
            ]

            logger.info(f"Deleting {len(template_ids)} templates from Zabbix")

            if len(template_ids):
                if not Settings.DRY_RUN:
                    self._zabbix.delete_templates(template_ids)

        return deletion_queue



import unittest
from os import getenv

from base_icon_map import BaseIconMap

DEV_ZABBIX_URL = getenv("ZABBIX_URL")
DEV_ZABBIX_TOKEN = getenv("ZABBIX_TOKEN")
DEV_GIT_REMOTE = getenv("REMOTE")


class TestIconMap(BaseIconMap, unittest.IsolatedAsyncioTestCase):
    pass
